#include "MainWindow.h"
#include "ui_MainWindow.h"
#include <QSerialPortInfo>
#include <QMessageBox>
#include <QDateTime>
#include <fstream>
#include <QFileDialog>


MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow) {
    serial = nullptr;
    ui->setupUi(this);
    ui->log->document()->setTextWidth(20);
    ui->log->setReadOnly(true);
    setWindowTitle("Podval Sniffer soft");
    updatePortsList();
}

MainWindow::~MainWindow() {
    delete serial;
    delete ui;
}

void MainWindow::on_updatePortsButton_clicked() {
    updatePortsList();
}

void MainWindow::on_connectButton_clicked() {
    auto port = ui->portSelect->currentText();
    if (port.isEmpty()) {
        ui->log->appendHtml("<font style=\"color:red\">[ERR] Порт не выбран!</font>\n");
    } else {
        ui->log->appendHtml("<font style=\"color:green\">[INFO] Попытка открыть порт '" + port + "'</font>\n");
        if (serial) {
            serial->close();
        }
        serial = new QSerialPort(port);
        connect(serial, &QSerialPort::readyRead, this, &MainWindow::readDataHandler);

        serial->setBaudRate(ui->speedSelect->currentText().toInt());
        serial->setDataBits(QSerialPort::Data8);
        serial->setFlowControl(QSerialPort::NoFlowControl);
        serial->setStopBits(QSerialPort::OneStop);

        if (serial->open(QIODevice::ReadWrite)) {
            ui->log->appendHtml("<font style=\"color:green\">[INFO] Успех!</font>\n");
        } else {
            ui->log->appendHtml("<font style=\"color:red\">[ERR] Не удалось открыть порт!</font>\n");
            QMessageBox::critical(this, tr("Error"), serial->errorString());
        }
    }
}

void MainWindow::on_disconnect_clicked() {
    if (serial) {
        if (serial->isOpen()) {
            serial->close();
        }
        delete serial;
        serial = nullptr;
    }
    ui->log->appendHtml("<font style=\"color:green\">[INFO] Порт <font style=\"color:red\">отключен</font>!</font>\n");
}

void MainWindow::on_clearLog_clicked() {
    ui->log->clear();
    messages.clear();
}

static std::string messageStdIdToString(uint16_t stdId) {
    char groupBuff[64];
    if ((stdId & 0x7C0) == 0x7C0) {
        // group = 4;
        sprintf(groupBuff, "(%03X) Group 4, MsgId=0x%02X             ", stdId, (stdId & 0x3F));
    } else if ((stdId & 0x600) == 0x600) {
        // group = 3;
        sprintf(groupBuff, "(%03X) Group 3, MsgId=0x%02X, SrcMAC=0x%02X",
                stdId, ((stdId & 0x1C0) >> 6), (stdId & 0x3F));
    } else if ((stdId & 0x600) == 0x400) {
        // group = 2;
        sprintf(groupBuff, "(%03X) Group 2, MsgId=0x%02X, MAC-ID=0x%02X",
                stdId, (stdId & 0x7), ((stdId & 0x1F8) >> 3));
    } else if ((stdId & 0x400) == 0) {
        // group = 1;
        sprintf(groupBuff, "(%03X) Group 1, MsgId=0x%02X, SrcMAC=0x%02X",
                stdId, ((stdId & 0x3C0) >> 6), (stdId & 0x3F));
    } else {
        sprintf(groupBuff, "(%03X) Group is incorrect!", stdId);
    }
    return groupBuff;
}

void MainWindow::on_storeToFile_clicked() {
    QString path = "sniff-" + QDateTime::currentDateTime().toString("yyyy.MM.dd-hh:mm:ss") + ".bin";
    std::fstream f(path.toStdString(), std::ios_base::out | std::ios_base::binary);
    if (f.good()) {
        for (const auto m : messages) {
            f.write((const char*) &m, sizeof(m));
        }
    }
    f.close();
    ui->log->appendHtml("<font style=\"color:green\">[INFO] Файл '" + path +
                        "' записан. Там " + QString::number(messages.size()) + " записей</font>\n");
}

void MainWindow::on_loadFromFile_clicked() {
    messages.clear();
    auto path = QFileDialog::getOpenFileName(this, "Open Dialog", "", "*.bin");
    if (!path.isEmpty()) {
        std::fstream f(path.toStdString(), std::ios_base::in | std::ios_base::binary);
        if (f.good()) {
            do {
                MessageStruct m{};
                f.read((char*) &m, sizeof(m));
                if (f.good()) {
                    messages.push_back(m);
                } else {
                    break;
                }
            } while (!f.eof());
            ui->log->appendHtml("<font style=\"color:green\">[INFO] Файл '" + path +
                                "' прочитан. Там " + QString::number(messages.size()) + " записей</font>\n");
        }
        f.close();
    }
}

void MainWindow::on_generateTextFile_clicked() {
    QString path = "sniff-" + QDateTime::currentDateTime().toString("yyyy.MM.dd-hh:mm:ss") + ".txt";
    std::fstream f(path.toStdString(), std::ios_base::out);
    if (f.good()) {
        for (const auto m : messages) {
            std::string text = messageToString(m).toStdString() + "\n";
            f.write(text.c_str(), text.size());
        }
    }
    f.close();
    ui->log->appendHtml("<font style=\"color:green\">[INFO] Файл '" + path +
                        "' записан. Там " + QString::number(messages.size()) + " записей</font>\n");
}

QString MainWindow::messageToString(const MessageStruct& msg) {
    char buff[1024];
    std::string time = QDateTime::fromMSecsSinceEpoch((qint64) msg.milliseconds).
            toString("yyyy.MM.dd hh:mm:ss.zzz").toStdString();
    MessageDecoder decoder(msg);
    sprintf(buff, "%s\n[%s]: %38s, DataLen=%X, Data={%02X %02X %02X %02X %02X %02X %02X %02X}\n",
            decoder.toString().c_str(), time.c_str(), messageStdIdToString(msg.stdId).c_str(), msg.dataLen,
            msg.data[0], msg.data[1], msg.data[2], msg.data[3], msg.data[4], msg.data[5], msg.data[6], msg.data[7]);
    return {buff};
}

static uint8_t charToNum(char data) {
    if (data >= '0' && data <= '9') return data - '0';
    else if (data >= 'a' && data <= 'f') return data - 'a' + 10;
    else if (data >= 'A' && data <= 'F') return data - 'A' + 10;
    else return 0;
}

void MainWindow::putCharToMessage(char data) {
    if (data == '>') {
        // начало посылки
        automat.message.milliseconds = QDateTime::currentMSecsSinceEpoch();
        automat.currPointer = 0;
        automat.discard = false;
    } else if (data == '\n') {
        // конец посылки
        if (!automat.discard) {
            // обработка пакетов
            if (automat.currPointer != 20) {
                // ошибка
                ui->log->appendHtml("[ERR] *** packet crush ***!\n");
            } else {
                // вот тут дрочильня с пакетом
                // сначала парсим StdId
                automat.message.stdId = charToNum(automat.text[2]) |
                           (charToNum(automat.text[1]) << 4) |
                           (charToNum(automat.text[0]) << 8);

                // теперь длинну посылки
                automat.message.dataLen = charToNum(automat.text[3]);

                // теперь саму посылку, она все равно всегда передается как 8 байт
                for (int i = 0; i < 8; i++) {
                    automat.message.data[i] = charToNum(automat.text[(i << 1) + 5]) | (charToNum(automat.text[(i << 1) + 4]) << 4);
                }

                messages.push_back(automat.message);
                ui->log->appendPlainText(messageToString(automat.message));

//                printf(">%03X%X%02X%02X%02X%02X%02X%02X%02X%02X\n",
//                       automat.message.stdId, automat.message.dataLen,
//                       automat.message.data[0], automat.message.data[1], automat.message.data[2], automat.message.data[3],
//                       automat.message.data[4], automat.message.data[5], automat.message.data[6], automat.message.data[7]);
            }
        }

        automat.discard = true; // так, на всякий случай
    } else {
        if (!automat.discard) {
            if (automat.currPointer != 20) {
                automat.text[automat.currPointer++] = data;
            } else {
                automat.discard = true;
            }
        }
    }
}

void MainWindow::readDataHandler() {
    if (serial) {
        auto array = serial->readAll();
        for (char i : array) {
            putCharToMessage(i);
        }
    }
}

void MainWindow::updatePortsList() {
    auto ports = QSerialPortInfo::availablePorts();

    ui->portSelect->clear();
    for (const auto& p : ports) {
        ui->portSelect->addItem(p.portName());
    }
}

